/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2010                                                    *
 * Copyright (C) 1998-2010 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef PARAMENGINE_H_
#define PARAMENGINE_H_

#include <string>
#include <list>
#include <iostream>
#include <cstring>
#include "Utilities.h"

/**
 * ͣ
 * Ϊ˼ȥ V_BOOL ͣV_BOOL ʾѡͲģ
 *  -name ʾ name Ϊ true -goalie, -dd,
 * ҪʵͬĹΪ -goalie on, -dd on
 */
enum ParamType {
	V_NONE,

	V_INT,		//
	V_DOUBLE,	//ʵ
	V_STRING,	//ַ -- ֻ std::string
	V_ONOFF		//
};

template < typename T >
struct ParamTypes {
	static const ParamType TYPE = V_NONE;
};

template <>
struct ParamTypes <int>
{
	static const ParamType TYPE = V_INT;
};

template <>
struct ParamTypes <double>
{
	static const ParamType TYPE = V_DOUBLE;
};

template <>
struct ParamTypes <std::string>
{
	static const ParamType TYPE = V_STRING;
};

template <>
struct ParamTypes <bool>
{
	static const ParamType TYPE = V_ONOFF;
};

struct Param {
	std::string name;
	void *ptr;        //pointer to data field
	ParamType type;

	Param(const char *str, void *p, ParamType t):
		name(str),
		ptr(p),
		type(t)
	{
	}
};

/**
 * Ҫֵ/serverϢ/ļ;
 */
class ParamEngine {
public:
	typedef std::list<Param>::iterator ParamPtr;

private:
	typedef std::list<Param> ParamList;

	enum {
		HASH_SIZE = 512
	};

	Array<ParamList, HASH_SIZE> mParamLists;

private:
	int hash(const char *str) {
		if (str != 0)
		{
			unsigned h = str [0];
			for (int i = 1; str [i] != 0; ++i)
			{
				h = (h << 5) - h + str [i];
			}
			return h % HASH_SIZE; // remainder
		}
		else
		{
			return -1;
		}
	}

	bool str_equal(const char *s1, const char *b) {
		return !strcmp(s1, b);
	}

public:
	ParamEngine() {}
	virtual ~ParamEngine() {}

	template <typename T>
	void AddParam(const char *name, T *address, T value){
		if (address){
			// assign default value
			*address = value;

			// allocate Param struct
			Param param(name, static_cast<void*>(address), ParamTypes<T>::TYPE);

			if (param.type == V_NONE) {
				std::cerr << __FILE__ << ":" << __LINE__ << ": param type error \"" << name << "\"" << std::endl;
				return;
			}
			// put into list
			int id = hash(name);
			if (id > -1){
				this->mParamLists[id].push_back(param);
			}
		}
	}

	template <typename T>
	bool SetParam(const char *name, T value)
	{
		if (name != 0 ){
			ParamPtr it;
			bool found = GetParam(name, it);
			if (found == true){
				switch(it->type){
					case V_INT:
					case V_DOUBLE:
					case V_STRING:
					case V_ONOFF:
						*static_cast<T*>(it->ptr) = value;
						break;
					default:
						return false;
				}
				return true;
			}
			else {
				std::cerr << __FILE__ << ":" << __LINE__ << ": unknown argument \'" << name << "\'" << std::endl;
			}
		}
		return false;
	}

	virtual void AddParams() = 0;

	bool SetParamFromString(const char *name, const char *buffer);

	bool GetParam(const char *name, ParamPtr & it);
	bool PrintParam(const char *name, std::ostream &os);
	bool PrintParam(const char *name);
	void DumpParam(std::ostream &os);
	void DumpParam();
	bool SaveToConfigFile(const char *file_name);

	/**
	 * ȡϢʽ-name value ...
	 * @param argc 
	 * @param argv ַ׵ַ
	 */
	void ParseFromCmdLine(int argc, char **argv);

	/**
	 * serverϢȡϢʽ(name value) ...
	 * @param line serverϢԭ
	 */
	void ParseFromServerMsg(const char *line);

	/**
	 * ļȡϢʽname = value
	 * @param file_name
	 * @return Ƿɹȡ
	 */
	bool ParseFromConfigFile(const char *file_name);
};

#endif /* PARAMENGINE_H_ */
