/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2010                                                    *
 * Copyright (C) 1998-2010 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef __TimeTest_H__
#define __TimeTest_H__


#include <vector>
#include <cstring>
#include "Utilities.h"

/**
 * ԺʱĽӿ
 * Interface to calculate the time cost of a function.
 */
#define TIMETEST(func_name) TimeTestFunc time_test_func(func_name);


/**
 * TimeCost.
 */
struct TimeCost
{
    /**
     * Constructor.
     */
    TimeCost()
    {
        mNum        = 0;
        mAveCost    = 0;
        mMaxTime    = Time(-3, 0);
        mMaxCost    = 0;
        mMinTime    = Time(-3, 0);
        mMinCost    = 1000000;
    }

	/**
	 * 
	 * Number of times a function was called.
	 */
    long    mNum;

    /**
     * ƽʱ
     * Average cost.
     */
    double  mAveCost; //

    /**
     * ʱӦ
     * Cycle which the highest cost was generated at.
     */
    Time    mMaxTime;

    /**
     * ʱ
     * Highest time cost.
     */
    long    mMaxCost;

    /**
     * СʱӦ
     * Cycle which the lowest cost was generated at.
     */
    Time    mMinTime;

    /**
     * Сʱ
     * Lowest time cost.
     */
    long    mMinCost;
};


/**
 * TimeRecord.
 */
struct TimeRecord
{
	/**
	 * Constructor.
	 */
    TimeRecord()
    {
        mBeginTime      = RealTime(0, 0);
        mCycleTimeCost  = 0;
    }

    /**
	 * ÿڵϢ
	 * Time cost for every cycle.
	 */
    TimeCost    mEachCycle;

    /**
     * ÿһεϢ
     * Time cost for each call.
     */
    TimeCost    mEachTime;

    /**
     * ÿһεĿʼʱ
     * Begin time for each call.
     */
    RealTime    mBeginTime;

    /**
     * ÿһܹʱ
     * Total cost for each cycle.
     */
    long        mCycleTimeCost;
};


/**
 * TimeTest.
 */
class TimeTest
{
    TimeTest();

public:
    ~TimeTest();

    /**
     * ʵ
     * Instacne.
     */
    static TimeTest & instance();

    /**
     * µڵи£current_timeΪǰǸ
     * Update when each cycle begins.
     */
    void Update(Time current_time);

    /**
     * ÿβԵĿʼ
     * Begin of each test.
     */
    int Begin(std::string func_name);

    /**
     * ÿβԵĽ
     * End of each test.
     */
    void End(int event_id);

    /**
     * Լĺ룬ļĸֵ
     * Set self unum, which record files will be named after.
     */
    inline void SetUnum(int unum) { mUnum = unum; }

private:
    std::vector<std::string>    mEventQueue; // ¼Ժ
    std::vector<TimeRecord>     mRecordQueue; // ¼ԵĽ
    std::vector<bool>           mIsExecute; // ¼ǰǷִй
    std::vector<bool>           mIsBegin; // ֤Begin()End()ִ

    Time        mUpdateTime; // ϴupdate
    int         mUnum; // Լĺ
};


/**
 * TimeTestFunc
 */
class TimeTestFunc
{
public:
    TimeTestFunc(std::string func_name);
    ~TimeTestFunc();

private:
    int mEventID;
};


#endif

